import datetime
import unittest

import requests_mock
from pandas import DataFrame

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import Query, QueryType, UserBrief, QueryDefinition, FilterDefinition, FilterType, Database, \
    ContainerBrief, ImageBrief, Image, Table, Constraints, PrimaryKey, TableBrief, ColumnBrief, ColumnType, \
    Column, Operator, JoinType, JoinDefinition, ConditionalDefinition
from dbrepo.api.exceptions import MalformedError, NotExistsError, ForbiddenError, QueryStoreError, FormatNotAvailable, \
    ServiceError, ResponseCodeError, AuthenticationError


class QueryUnitTest(unittest.TestCase):
    database = Database(
        id="6bd39359-b154-456d-b9c2-caa516a45732",
        name='test',
        owner=UserBrief(id='abdbf897-e599-4e5a-a3f0-7529884ea011', username='mweise'),
        contact=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
        exchange_name='dbrepo',
        internal_name='test_abcd',
        is_public=True,
        is_schema_public=True,
        is_dashboard_enabled=True,
        tables=[Table(id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                      name="Some Table",
                      description="Test Table",
                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                      internal_name="some_table",
                      owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                      is_versioned=True,
                      queue_name='dbrepo',
                      routing_key='dbrepo.test_database_1234.some_table',
                      is_public=True,
                      is_schema_public=True,
                      constraints=Constraints(uniques=[],
                                              foreign_keys=[],
                                              checks=[],
                                              primary_key=[PrimaryKey(id="1516310f-ecb5-4614-abe2-3b96114e1484",
                                                                      table=TableBrief(
                                                                          id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='Some Table',
                                                                          internal_name='some_table',
                                                                          is_versioned=True,
                                                                          is_public=True,
                                                                          is_schema_public=True,
                                                                          owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16'),
                                                                      column=ColumnBrief(
                                                                          id="31a533b6-8ddf-43d6-ac6a-b9da597cb976",
                                                                          table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='id',
                                                                          alias=None,
                                                                          internal_name='id',
                                                                          type=ColumnType.BIGINT))]),
                      columns=[Column(id="31a533b6-8ddf-43d6-ac6a-b9da597cb976",
                                      name="ID",
                                      ord=0,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                      internal_name="id",
                                      type=ColumnType.BIGINT,
                                      is_null_allowed=False),
                               Column(id="85de93a8-834c-4cf4-9d34-f80ebd97e606",
                                      name="Username",
                                      ord=1,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="029d773f-f98b-40c0-ab22-b8b1635d4fbc",
                                      internal_name="username",
                                      type=ColumnType.VARCHAR,
                                      is_null_allowed=False)
                               ]),
                Table(id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                      name="Other Table",
                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                      internal_name="other_table",
                      owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                      is_versioned=True,
                      queue_name='dbrepo',
                      routing_key='dbrepo.test_database_1234.other_table',
                      is_public=True,
                      is_schema_public=True,
                      constraints=Constraints(uniques=[],
                                              foreign_keys=[],
                                              checks=[],
                                              primary_key=[PrimaryKey(id="3b060596-38f9-4055-8fb0-5526918f31a0",
                                                                      table=TableBrief(
                                                                          id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='Other Table',
                                                                          internal_name='other_table',
                                                                          is_versioned=True,
                                                                          is_public=True,
                                                                          is_schema_public=True,
                                                                          owned_by='8638c043-5145-4be8-a3e4-4b79991b0a16'),
                                                                      column=ColumnBrief(
                                                                          id="4a2f20c3-9efd-4788-a38d-3079d198125c",
                                                                          table_id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                                                                          database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                                          name='id',
                                                                          alias=None,
                                                                          internal_name='id',
                                                                          type=ColumnType.BIGINT))]),
                      columns=[Column(id="4a2f20c3-9efd-4788-a38d-3079d198125c",
                                      name="ID",
                                      ord=0,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                                      internal_name="id",
                                      type=ColumnType.BIGINT,
                                      is_null_allowed=False),
                               Column(id="85de93a8-834c-4cf4-9d34-f80ebd97e606",
                                      name="City",
                                      ord=1,
                                      database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                      table_id="585d421a-ad1a-4543-b661-5e32a78dd3e1",
                                      internal_name="city",
                                      type=ColumnType.VARCHAR,
                                      is_null_allowed=False)
                               ])
                ],
        container=ContainerBrief(id="44d811a8-4019-46ba-bd57-ea10a2eb0c74",
                                 name='MariaDB Galera 11.1.3',
                                 internal_name='mariadb',
                                 image=ImageBrief(id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                                                  name='mariadb',
                                                  version='11.2.2',
                                                  default=True)))
    image = Image(id="b104648b-54d2-4d72-9834-8e0e6d428b39",
                  name='mariadb',
                  version='11.2.2',
                  default=True,
                  data_types=[],
                  operators=[Operator(id="6a96bd99-be3d-4d56-8c38-b14bdfead634",
                                      display_name="IN",
                                      value="IN",
                                      documentation="https://mariadb.com/kb/en/in/")])

    def test_create_subset_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(exp)
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', json=exp,
                      headers={'X-Id': '85bc1217-29ab-4c09-9f98-8c019238a9c8', 'X-Headers': 'id,username'},
                      status_code=201)
            # test
            client = RestClient(username="a", password="b")
            response = client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732", page=0, size=10,
                                            timestamp=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                                            query=QueryDefinition(datasources=["some_table"],
                                                                  columns=["some_table.id", "some_table.username"],
                                                                  filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                            column="some_table.id",
                                                                                            operator="IN",
                                                                                            value="(1,2)")]))
            self.assertTrue(DataFrame.equals(df, response))

    def test_create_subset_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     query=QueryDefinition(datasources=["some_table"],
                                                           columns=["some_table.id", "some_table.username"],
                                                           filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                     column="some_table.id",
                                                                                     operator="IN",
                                                                                     value="(1,2)")]))
            except MalformedError:
                pass

    def test_create_subset_notion_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     query=QueryDefinition(datasources=["some_table"],
                                                           columns=["id",  # <<<
                                                                    "some_table.username"]))
            except MalformedError:
                pass

    def test_create_subset_notion_join_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     query=QueryDefinition(datasources=["some_table"],
                                                           columns=["some_table.id", "some_table.username",
                                                                    "other_table.city"],
                                                           joins=[JoinDefinition(type=JoinType.INNER,
                                                                                 datasource="other_table",
                                                                                 conditionals=[ConditionalDefinition(
                                                                                     column="username",  # <<<
                                                                                     foreign_column="other_table.username")])]))
            except MalformedError:
                pass

    def test_create_subset_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     query=QueryDefinition(datasources=["some_table"],
                                                           columns=["some_table.id", "some_table.username"],
                                                           filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                     column="some_table.id",
                                                                                     operator="IN",
                                                                                     value="(1,2)")]))
            except ForbiddenError:
                pass

    def test_create_subset_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     query=QueryDefinition(datasources=["some_table"],
                                                           columns=["some_table.id", "some_table.username"],
                                                           filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                     column="some_table.id",
                                                                                     operator="IN",
                                                                                     value="(1,2)")]))
            except NotExistsError:
                pass

    def test_create_subset_417_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', status_code=417)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     query=QueryDefinition(datasources=["some_table"],
                                                           columns=["some_table.id", "some_table.username"],
                                                           filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                     column="some_table.id",
                                                                                     operator="IN",
                                                                                     value="(1,2)")]))
            except QueryStoreError:
                pass

    def test_create_subset_501_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', status_code=501)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     query=QueryDefinition(datasources=["some_table"],
                                                           columns=["some_table.id", "some_table.username"],
                                                           filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                     column="some_table.id",
                                                                                     operator="IN",
                                                                                     value="(1,2)")]))
            except FormatNotAvailable:
                pass

    def test_create_subset_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     query=QueryDefinition(datasources=["some_table"],
                                                           columns=["some_table.id", "some_table.username"],
                                                           filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                     column="some_table.id",
                                                                                     operator="IN",
                                                                                     value="(1,2)")]))
            except ServiceError:
                pass

    def test_create_subset_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                     query=QueryDefinition(datasources=["some_table"],
                                                           columns=["some_table.id", "some_table.username"],
                                                           filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                     column="some_table.id",
                                                                                     operator="IN",
                                                                                     value="(1,2)")]))
            except ResponseCodeError:
                pass

    def test_create_subset_anonymous_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(exp)
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', json=exp,
                      headers={'X-Id': '85bc1217-29ab-4c09-9f98-8c019238a9c8', 'X-Headers': 'id,username'},
                      status_code=201)
            # test

            client = RestClient()
            response = client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732", page=0, size=10,
                                            query=QueryDefinition(datasources=["some_table"],
                                                                  columns=["some_table.id", "some_table.username"],
                                                                  filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                            column="some_table.id",
                                                                                            operator="IN",
                                                                                            value="(1,2)")]))
            self.assertTrue(DataFrame.equals(df, response))

    def test_create_subset_alias_anonymous_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(exp)
            # mock
            mock.get(f'/api/v1/image/{self.image.id}', json=self.image.model_dump(),
                     status_code=200)
            mock.get(f'/api/v1/database/{self.database.id}', json=self.database.model_dump(),
                     status_code=200)
            mock.post(f'/api/v1/database/{self.database.id}/subset', json=exp,
                      headers={'X-Id': '85bc1217-29ab-4c09-9f98-8c019238a9c8', 'X-Headers': 'id,username'},
                      status_code=201)
            # test
            try:
                client = RestClient()
                response = client.create_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732", page=0, size=10,
                                                query=QueryDefinition(datasources=["some_table"],
                                                                      joins=[JoinDefinition(type=JoinType.INNER,
                                                                                            datasource="other_table",
                                                                                            conditionals=[
                                                                                                ConditionalDefinition(
                                                                                                    column="username",
                                                                                                    foreign_column="username")])],
                                                                      columns=["some_table.id", "some_table.username",
                                                                               "other_table.city"],
                                                                      filters=[FilterDefinition(type=FilterType.WHERE,
                                                                                                column="some_table.id",
                                                                                                operator="IN",
                                                                                                value="(1,2)")]))
            except MalformedError:
                pass

    def test_get_subset_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Query(id="e1df2bb8-1f12-494a-ade5-2c4aecdab939",
                        owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                        execution=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                        query='SELECT id, username FROM some_table WHERE id IN (1,2)',
                        query_normalized='SELECT id, username FROM some_table WHERE id IN (1,2)',
                        type=QueryType.QUERY,
                        database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                        query_hash='da5ff66c4a57683171e2ffcec25298ee684680d1e03633cd286f9067d6924ad8',
                        result_hash='464740ba612225913bb15b26f13377707949b55e65288e89c3f8b4c6469aecb4',
                        is_persisted=False,
                        result_number=None,
                        identifiers=[])
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                headers={'X-Headers': 'id,username'}, json=exp.model_dump())
            # test
            response = RestClient().get_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                               subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            self.assertEqual(exp, response)

    def test_get_subset_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=403)
            # test
            try:
                RestClient().get_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                        subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except ForbiddenError:
                pass

    def test_get_subset_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=404)
            # test
            try:
                RestClient().get_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                        subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except NotExistsError:
                pass

    def test_get_subset_406_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=406)
            # test
            try:
                RestClient().get_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                        subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except FormatNotAvailable:
                pass

    def test_get_subset_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=503)
            # test
            try:
                RestClient().get_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                        subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except ServiceError:
                pass

    def test_get_subset_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=202)
            # test
            try:
                RestClient().get_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                        subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except ResponseCodeError:
                pass

    def test_get_queries_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = []
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/subset', headers={'X-Headers': ''}, json=[])
            # test
            response = RestClient().get_queries(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            self.assertEqual(exp, response)

    def test_update_subset_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Query(id="e1df2bb8-1f12-494a-ade5-2c4aecdab939",
                        owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                        execution=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                        query='SELECT id, username FROM some_table WHERE id IN (1,2)',
                        query_normalized='SELECT id, username FROM some_table WHERE id IN (1,2)',
                        type=QueryType.QUERY,
                        database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                        query_hash='da5ff66c4a57683171e2ffcec25298ee684680d1e03633cd286f9067d6924ad8',
                        result_hash='464740ba612225913bb15b26f13377707949b55e65288e89c3f8b4c6469aecb4',
                        is_persisted=True,
                        result_number=None,
                        identifiers=[])
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                json=exp.model_dump(), headers={'X-Headers': 'id,username'}, status_code=202)
            # test
            response = RestClient(username='foo', password='bar').update_subset(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732", subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939",
                persist=True)
            self.assertEqual(exp, response)

    def test_update_subset_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=400)
            # test
            try:
                RestClient(username='foo', password='bar').update_subset(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939", persist=True)
            except MalformedError:
                pass

    def test_update_subset_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=403)
            # test
            try:
                RestClient(username='foo', password='bar').update_subset(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939", persist=True)
            except ForbiddenError:
                pass

    def test_update_subset_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=404)
            # test
            try:
                RestClient(username='foo', password='bar').update_subset(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939", persist=True)
            except NotExistsError:
                pass

    def test_update_subset_417_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=417)
            # test
            try:
                RestClient(username='foo', password='bar').update_subset(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939", persist=True)
            except QueryStoreError:
                pass

    def test_update_subset_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=503)
            # test
            try:
                RestClient(username='foo', password='bar').update_subset(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939", persist=True)
            except ServiceError:
                pass

    def test_update_subset_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939',
                status_code=200)
            # test
            try:
                RestClient(username='foo', password='bar').update_subset(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                    subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939",
                    persist=True)
            except ResponseCodeError:
                pass

    def test_update_subset_anonymous_fails(self):
        # test
        try:
            RestClient().update_subset(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                       subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939",
                                       persist=True)
        except AuthenticationError:
            pass

    def test_get_queries_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [Query(id="e1df2bb8-1f12-494a-ade5-2c4aecdab939",
                         owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'),
                         execution=datetime.datetime(2024, 1, 1, 0, 0, 0, 0, datetime.timezone.utc),
                         query='SELECT id, username FROM some_table WHERE id IN (1,2)',
                         query_normalized='SELECT id, username FROM some_table WHERE id IN (1,2)',
                         type=QueryType.QUERY,
                         database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                         query_hash='da5ff66c4a57683171e2ffcec25298ee684680d1e03633cd286f9067d6924ad8',
                         result_hash='464740ba612225913bb15b26f13377707949b55e65288e89c3f8b4c6469aecb4',
                         is_persisted=False,
                         result_number=None,
                         identifiers=[])]
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/subset', headers={'X-Headers': 'id,username'},
                     json=[exp[0].model_dump()])
            # test
            response = RestClient().get_queries(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            self.assertEqual(exp, response)

    def test_get_queries_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/subset', status_code=403)
            # test
            try:
                RestClient().get_queries(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ForbiddenError:
                pass

    def test_get_queries_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/subset', status_code=404)
            # test
            try:
                RestClient().get_queries(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except NotExistsError:
                pass

    def test_get_queries_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/subset', status_code=503)
            # test
            try:
                RestClient().get_queries(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ServiceError:
                pass

    def test_get_queries_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/database/{self.database.id}/subset', status_code=202)
            # test
            try:
                RestClient().get_queries(database_id="6bd39359-b154-456d-b9c2-caa516a45732")
            except ResponseCodeError:
                pass

    def test_get_subset_data_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(exp)
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                headers={'X-Headers': 'id,username'}, json=exp)
            # test
            response = RestClient().get_subset_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                    subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_subset_data_dataframe_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [{'id': 1, 'username': 'foo'}, {'id': 2, 'username': 'bar'}]
            df = DataFrame.from_records(exp)
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                headers={'X-Headers': 'id,username'}, json=exp)
            # test
            response = RestClient().get_subset_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                    subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            self.assertEqual(df.shape, response.shape)
            self.assertTrue(DataFrame.equals(df, response))

    def test_get_subset_data_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=400)
            # test
            try:
                RestClient().get_subset_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                             subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except MalformedError:
                pass

    def test_get_subset_data_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=403)
            # test
            try:
                RestClient().get_subset_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                             subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except ForbiddenError:
                pass

    def test_get_subset_data_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=404)
            # test
            try:
                RestClient().get_subset_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                             subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except NotExistsError:
                pass

    def test_get_subset_data_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=503)
            # test
            try:
                RestClient().get_subset_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                             subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except ServiceError:
                pass

    def test_get_subset_data_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=202)
            # test
            try:
                RestClient().get_subset_data(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                             subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except ResponseCodeError:
                pass

    def test_get_subset_data_count_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = 2
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                headers={'X-Count': str(exp)})
            # test
            response = RestClient().get_subset_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                          subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            self.assertEqual(exp, response)

    def test_get_subset_data_count_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=400)
            # test
            try:
                RestClient().get_subset_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                   subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except MalformedError:
                pass

    def test_get_subset_data_count_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=403)
            # test
            try:
                RestClient().get_subset_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                   subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except ForbiddenError:
                pass

    def test_get_subset_data_count_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=404)
            # test
            try:
                RestClient().get_subset_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                   subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except NotExistsError:
                pass

    def test_get_subset_data_count_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=503)
            # test
            try:
                RestClient().get_subset_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                   subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except ServiceError:
                pass

    def test_get_subset_data_count_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.head(
                '/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/subset/e1df2bb8-1f12-494a-ade5-2c4aecdab939/data',
                status_code=202)
            # test
            try:
                RestClient().get_subset_data_count(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                   subset_id="e1df2bb8-1f12-494a-ade5-2c4aecdab939")
            except ResponseCodeError:
                pass


if __name__ == "__main__":
    unittest.main()



import os
import boto3
import logging

from flask import current_app
from boto3.exceptions import S3UploadFailedError
from botocore.exceptions import ClientError


class StorageServiceClient:

    def __init__(self, endpoint: str, access_key_id: str, secret_access_key: str):
        endpoint = endpoint
        aws_access_key_id = access_key_id
        aws_secret_access_key = secret_access_key
        logging.info(
            f"retrieve file from S3, endpoint={current_app.config['S3_PROTO']}://{endpoint}, aws_access_key_id={aws_access_key_id}, aws_secret_access_key=(hidden)")
        self.client = boto3.client(service_name='s3', endpoint_url=f"{current_app.config['S3_PROTO']}://{endpoint}",
                                   aws_access_key_id=aws_access_key_id,
                                   aws_secret_access_key=aws_secret_access_key)
        self.bucket_exists_or_exit(current_app.config['S3_BUCKET'])

    def upload_file(self, filename: str, path: str = "/tmp", bucket: str = "dbrepo") -> bool:
        """
        Uploads a file to the blob storage.
        Follows the official API https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-uploading-files.html.
        :param bucket: The bucket to upload the file into.
        :param path: The path the file is located.
        :param filename: The filename.
        :return: True if the file was uploaded.
        """
        filepath = os.path.join(path, filename)
        if os.path.isfile(filepath):
            logging.info(f'Found .csv at {filepath}')
        else:
            logging.error(f'Failed to find .csv at {filepath}')
            raise FileNotFoundError(f'Failed to find .csv at {filepath}')
        try:
            self.client.upload_file(filepath, bucket, filename)
            logging.info(f"Uploaded .csv {filepath} with key {filename}")
            return True
        except (ClientError, S3UploadFailedError) as e:
            logging.warning(f"Failed to upload file with key {filename}")
            raise ConnectionRefusedError(f"Failed to upload file with key {filename}", e)

    def download_file(self, filename: str, path: str = "/tmp", bucket: str = "dbrepo"):
        """
        Downloads a file from the blob storage.
        Follows the official API https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-example-download-file.html
        :param filename: The filename.
        :param path: The path the file is located.
        :param bucket: The bucket to download the file from.
        """
        self.file_exists(bucket, filename)
        filepath = os.path.join(path, filename)
        self.client.download_file(bucket, filename, filepath)
        logging.info(f"Downloaded .csv with key {filename} into {filepath}")

    def file_exists(self, bucket, filename):
        try:
            self.client.head_object(Bucket=bucket, Key=filename)
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                logging.error("Failed to find key %s in bucket %s", filename, bucket)
            else:
                logging.error("Unexpected error when finding key %s in bucket %s: %s", filename, bucket,
                              e.response["Error"]["Code"])
            raise e

    def get_file(self, bucket, filename):
        try:
            return self.client.get_object(Bucket=bucket, Key=filename)
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                logging.error("Failed to get file with key %s in bucket %s", filename, bucket)
            else:
                logging.error("Unexpected error when get file with key %s in bucket %s: %s", filename, bucket,
                              e.response["Error"]["Code"])
            raise e

    def bucket_exists_or_exit(self, bucket):
        try:
            return self.client.head_bucket(Bucket=bucket)
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                raise FileNotFoundError(f"Failed to find bucket: {bucket}")
            raise ConnectionError(f"Unexpected error when finding bucket {bucket}: %s", e.response["Error"]["Code"])



class UnitConversionException(Exception):

    def __init__(self, message):
        self.message = message


class ScaleConversionException(Exception):

    def __init__(self, message):
        self.message = message



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import UnitBrief
from dbrepo.api.exceptions import ResponseCodeError


class UserUnitTest(unittest.TestCase):

    def test_get_units_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [UnitBrief(id="d8eec1ab-7b37-4eb3-bdf7-b44a8b384c0b",
                             uri='http://www.ontology-of-units-of-measure.org/resource/om-2/CelsiusTemperature',
                             name='Celsius Temperature')]
            # mock
            mock.get('/api/v1/unit', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_units()
            self.assertEqual(exp, response)

    def test_get_units_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/unit', status_code=202)
            # test
            try:
                RestClient().get_units()
            except ResponseCodeError:
                pass


if __name__ == "__main__":
    unittest.main()



import math

from dbrepo.core.omlib.constants import SI
from dbrepo.core.omlib.exceptions.dimensionexception import DimensionalException
from dbrepo.core.omlib.scale import Scale
from dbrepo.core.omlib.thing import Thing
from dbrepo.core.omlib.unit import Unit, PrefixedUnit, SingularUnit


def om(numerical_value, unit_or_scale, identifier=None):
    if isinstance(unit_or_scale, Unit):
        return Measure(numerical_value, unit_or_scale, identifier)
    if isinstance(unit_or_scale, Scale):
        return Point(numerical_value, unit_or_scale, identifier)
    return None


class Point(Thing):

    @staticmethod
    def create_by_converting(point, to_scale):
        if not isinstance(point, Point):
            raise ValueError("The parameter to the convert method is not of the correct type (Point).")
        if not isinstance(to_scale, Scale):
            raise ValueError("The parameter to the convert method is not of the correct type (Scale).")
        new_point = Point(point.numericalValue, point.scale)
        new_point.convert(to_scale)
        return new_point

    @staticmethod
    def create_by_converting_to_ratio_scale(point):
        if not isinstance(point, Point):
            raise ValueError("The parameter to the convert method is not of the correct type (Point).")
        new_point = Point(point.numericalValue, point.scale)
        new_point.convert_to_ratio_scale()
        return new_point

    def __init__(self, numerical_value, scale, identifier=None):
        super().__init__(identifier=identifier)
        self.numericalValue = numerical_value
        self.scale = scale

    def convert(self, to_scale):
        if not isinstance(to_scale, Scale):
            raise ValueError("The parameter to the convert method is not of the correct type (Scale).")
        factor = Scale.conversion_factor(self.scale, to_scale)
        off_set = Scale.conversion_off_set(self.scale, to_scale)
        self.numericalValue = self.numericalValue * factor + off_set
        self.scale = to_scale

    def convert_to_ratio_scale(self):
        base = self.scale.base_ratio_scale()
        factor = Scale.conversion_factor(self.scale, base[0])
        off_set = Scale.conversion_off_set(self.scale, base[0])
        self.numericalValue = self.numericalValue * factor + off_set
        self.scale = base[0]

    def __str__(self):
        return f'{self.numericalValue} {self.scale.unit.symbol()}'

    def __new_value_for_comparisson(self, other):
        if isinstance(other, Point):
            factor = Scale.conversion_factor(self.scale, other.scale)
            off_set = Scale.conversion_off_set(self.scale, other.scale)
            new_value = self.numericalValue * factor + off_set
            return new_value
        return None

    def __eq__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value == other.numericalValue
        return False

    def __ne__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value != other.numericalValue
        return False

    def __lt__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value < other.numericalValue
        return False

    def __le__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value <= other.numericalValue
        return False

    def __gt__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value > other.numericalValue
        return False

    def __ge__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value >= other.numericalValue
        return False

    def __add__(self, other):
        if not isinstance(other, Measure):
            raise ValueError('The value to be added is not a measure and only measures can be added to a point.')
        if not self.scale.dimensions == other.unit.dimensions:
            raise DimensionalException("Measures and Points with units of different dimensions cannot be added "
                                       "together. {} != {}"
                                       .format(self.scale.unit, other.unit))
        new_measure = Measure.create_by_converting(other, self.scale.unit)
        return_point = Point(self.numericalValue + new_measure.numericalValue, self.scale)
        return return_point

    def __sub__(self, other):
        if not isinstance(other, Measure) and not isinstance(other, Point):
            raise ValueError('The value to be subtracted is not a point or a measure and only measures or points '
                             'can be subtracted from a point.')
        if isinstance(other, Measure):
            if not self.scale.dimensions == other.unit.dimensions:
                raise DimensionalException("Measures and Points with units of different dimensions cannot be "
                                           "subtracted from each other. {} != {}".format(self.scale.unit, other.unit))
            new_measure = Measure.create_by_converting(other, self.scale.unit)
            return_point = Point(self.numericalValue - new_measure.numericalValue, self.scale)
            return return_point
        if isinstance(other, Point):
            if not self.scale.dimensions == other.scale.dimensions:
                raise DimensionalException("Measures and Points with units of different dimensions cannot be "
                                           "subtracted from each other. {} != {}".format(self.scale.unit, other.unit))
            new_point = Point.create_by_converting(other, self.scale)
            return_measure = Measure(self.numericalValue - new_point.numericalValue, self.scale.unit)
            return return_measure

    def __mul__(self, other):
        as_measure = Measure(self.numericalValue, self.scale.unit)
        return as_measure * other

    def __truediv__(self, other):
        as_measure = Measure(self.numericalValue, self.scale.unit)
        return as_measure / other


class Measure(Thing):

    @staticmethod
    def create_by_converting(measure, to_unit):
        if not isinstance(measure, Measure):
            raise ValueError("The parameter to the convert method is not of the correct type (Measure).")
        if not isinstance(to_unit, Unit):
            raise ValueError("The parameter to the convert method is not of the correct type (Unit).")
        new_measure = Measure(measure.numericalValue, measure.unit)
        new_measure.convert(to_unit)
        return new_measure

    @staticmethod
    def create_by_converting_to_base_units(measure, in_system_of_units=SI):
        if not isinstance(measure, Measure):
            raise ValueError("The parameter to the convert method is not of the correct type (Measure).")
        new_measure = Measure(measure.numericalValue, measure.unit)
        new_measure.convert_to_base_units(in_system_of_units)
        return new_measure

    @staticmethod
    def create_by_converting_to_convenient_units(measure, in_system_of_units=None, use_prefixes=True):
        if not isinstance(measure, Measure):
            raise ValueError("The parameter to the convert method is not of the correct type (Measure).")
        new_measure = Measure(measure.numericalValue, measure.unit)
        new_measure.convert_to_convenient_units(in_system_of_units, use_prefixes=use_prefixes)
        return new_measure

    def __init__(self, numerical_value, unit, identifier=None):
        super().__init__(identifier=identifier)
        self.numericalValue = numerical_value
        self.unit = unit

    def convert(self, to_unit):
        if not isinstance(to_unit, Unit):
            raise ValueError("The parameter to the convert method is not of the correct type (Unit).")
        factor = Unit.conversion_factor(self.unit, to_unit)
        self.numericalValue = self.numericalValue * factor
        self.unit = to_unit

    def convert_to_base_units(self, in_system_of_units=None):
        if in_system_of_units is None:
            in_system_of_units = self.unit.systemOfUnits
        if in_system_of_units is None:
            in_system_of_units = SI.SYSTEM_OF_UNITS
        base = Unit.get_base_units(self.unit, in_system_of_units)
        self.convert(base)

    def convert_to_convenient_units(self, system_of_units=None, use_prefixes=True):
        if system_of_units is None:
            system_of_units = self.unit.systemOfUnits
        test_units = self.unit.with_dimensions(self.unit.dimensions, in_system_of_units=system_of_units)
        selected_unit = self.unit
        log_selected_value = Measure.__get_log_value(self.numericalValue, selected_unit)
        for test_unit in test_units:
            if use_prefixes or not isinstance(test_unit, PrefixedUnit):
                factor = Unit.conversion_factor(self.unit, test_unit)
                value = abs(self.numericalValue * factor)
                log_value = Measure.__get_log_value(value, test_unit)
                if log_value < log_selected_value:
                    log_selected_value = log_value
                    selected_unit = test_unit
                if abs(log_value - log_selected_value) < 0.1 and isinstance(test_unit, SingularUnit):
                    log_selected_value = log_value
                    selected_unit = test_unit
                if abs(log_value - log_selected_value) < 0.1 and isinstance(test_unit, PrefixedUnit)\
                        and not isinstance(selected_unit, SingularUnit):
                    log_selected_value = log_value
                    selected_unit = test_unit
        self.convert(selected_unit)

    @staticmethod
    def __get_log_value(value, unit):
        if value == 0.0:
            return 0
        log_value = abs(math.log10(value))
        if value < 1:
            log_value = log_value + 2
        if not isinstance(unit, SingularUnit):
            log_value = log_value + 1
        return log_value

    def __str__(self):
        return f'{self.numericalValue} {self.unit.symbol()}'

    def __new_value_for_comparisson(self, other):
        if isinstance(other, Measure):
            factor = Unit.conversion_factor(self.unit, other.unit)
            new_value = self.numericalValue * factor
            return new_value
        return None

    def __eq__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value == other.numericalValue
        return False

    def __ne__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value != other.numericalValue
        return False

    def __lt__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value < other.numericalValue
        return False

    def __le__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value <= other.numericalValue
        return False

    def __gt__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value > other.numericalValue
        return False

    def __ge__(self, other):
        new_value = self.__new_value_for_comparisson(other)
        if new_value is not None:
            return new_value >= other.numericalValue
        return False

    def __add__(self, other):
        if not isinstance(other, Measure):
            raise ValueError('The value to be added is not a measure.')
        if not self.unit.dimensions == other.unit.dimensions:
            raise DimensionalException("Measures with units of different dimensions cannot be added together. {} != {}"
                                       .format(self.unit, other.unit))
        new_measure = Measure.create_by_converting(other, self.unit)
        return_measure = Measure(self.numericalValue + new_measure.numericalValue, self.unit)
        return return_measure

    def __sub__(self, other):
        if not isinstance(other, Measure):
            raise ValueError('The value to be subtracted is not a measure.')
        if not self.unit.dimensions == other.unit.dimensions:
            raise DimensionalException("Measures with units of different dimensions cannot be subtracted from each "
                                       "other. {} != {}".format(self.unit, other.unit))
        new_measure = Measure.create_by_converting(other, self.unit)
        return_measure = Measure(self.numericalValue - new_measure.numericalValue, self.unit)
        return return_measure

    def __mul__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            new_measure = Measure(self.numericalValue * other, self.unit)
            return new_measure
        if not isinstance(other, Measure) and not isinstance(other, Point):
            raise ValueError('The multiplicand is not a measure, a point, a float, or an int.')
        other_unit = None
        if isinstance(other, Measure):
            other_unit = other.unit
        if isinstance(other, Point):
            other_unit = other.scale.unit
        new_value = self.numericalValue * other.numericalValue
        new_unit = Unit.get_unit_multiplication(self.unit, other_unit)
        new_measure = Measure(new_value, new_unit)
        simple_unit = Unit.simplified_compound_unit(new_unit)
        new_measure.convert(simple_unit)
        return new_measure

    def __truediv__(self, other):
        if isinstance(other, float) or isinstance(other, int):
            new_measure = Measure(self.numericalValue / other, self.unit)
            return new_measure
        if not isinstance(other, Measure) and not isinstance(other, Point):
            raise ValueError('The denominator is not a measure, a point, a float, or an int.')
        other_unit = None
        if isinstance(other, Measure):
            other_unit = other.unit
        if isinstance(other, Point):
            other_unit = other.scale.unit
        new_value = self.numericalValue / other.numericalValue
        new_unit = Unit.get_unit_division(self.unit, other_unit)
        new_measure = Measure(new_value, new_unit)
        simple_unit = Unit.simplified_compound_unit(new_unit)
        new_measure.convert(simple_unit)
        return new_measure



from rdflib import BNode, URIRef, Literal, XSD


class Thing:

    def __init__(self, label=None, identifier=None):
        if identifier is None:
            self.identifier = BNode()
        else:
            self.identifier = URIRef(identifier)
        if label is None:
            self.prefLabels = []
        else:
            self.prefLabels = []
            self.__add_label_to_array(self.prefLabels, label, None)
        self.altLabels = []

    def __add_label_to_array(self, array, label, language=None):
        if isinstance(label, list):
            for item in label:
                self.__add_label_to_array(array, item, language)
        else:
            if language is None:
                if isinstance(label, Literal) and not label in array:
                    array.append(label)
                else:
                    label_lit = Literal(label, datatype=XSD.string)
                    if not label_lit in array:
                        array.append(label_lit)
            else:
                if isinstance(label, Literal) and not label in array:
                    label_lit = Literal(label.normalize, language)
                    array.append(label_lit)
                else:
                    label_lit = Literal(label, language)
                    if not label_lit in array:
                        array.append(label_lit)

    def add_preferred_label(self, label, language=None):
        # TODO Check if a preferred label in the specified language already exists, if so move the old one to alt labels
        self.__add_label_to_array(self.prefLabels, label, language)

    def add_alternative_label(self, label, language=None):
        self.__add_label_to_array(self.altLabels, label, language)

    def label(self):
        label = self.preferred_label()
        return label

    def preferred_label(self, language=None):
        result_label = None
        if language is None:
            for pref_label in self.prefLabels:
                if pref_label.language is None:
                    result_label = pref_label
            if result_label is None:
                for pref_label in self.prefLabels:
                    if pref_label.language == 'en':
                        result_label = pref_label
        else:
            for pref_label in self.prefLabels:
                if pref_label.language == language:
                    result_label = pref_label
        return result_label

    def all_labels(self):
        labels = []
        labels.extend(self.prefLabels)
        labels.extend(self.altLabels)
        return labels


class SymbolThing(Thing):

    def __init__(self, name=None, symbol=None, identifier=None):
        super().__init__(name, identifier)
        self.dimensions = []
        self.symbols = []
        self.add_symbol(symbol)

    def add_symbol(self, symbol, language=None):
        if isinstance(symbol, list):
            for item in symbol:
                self.__add_label_to_array(self.symbols, item, language)
        else:
            if language is None:
                if isinstance(symbol, Literal) and not symbol in self.symbols:
                    self.symbols.append(symbol)
                else:
                    label_lit = Literal(symbol, datatype=XSD.string)
                    if not label_lit in self.symbols:
                        self.symbols.append(label_lit)
            else:
                if isinstance(symbol, Literal) and not symbol in self.symbols:
                    label_lit = Literal(symbol.normalize, language)
                    self.symbols.append(label_lit)
                else:
                    label_lit = Literal(symbol, language)
                    if not label_lit in self.symbols:
                        self.symbols.append(label_lit)

    def symbol(self):
        symbol = self.preferred_symbol()
        return symbol

    def preferred_symbol(self, language=None):
        result_symbol = None
        if language is None:
            for symbol in self.symbols:
                if symbol.language is None:
                    result_symbol = symbol
            if result_symbol is None:
                for symbol in self.symbols:
                    if symbol.language == 'en':
                        result_symbol = symbol
        else:
            for symbol in self.symbols:
                if symbol.language == language:
                    result_symbol = symbol
        return result_symbol

    def all_symbols(self):
        return self.symbols



import unittest

from clients.s3_client import S3Client
from botocore.exceptions import ClientError


class StorageServiceClientIntegrationTest(unittest.TestCase):

    # @Test
    def test_upload_file_succeeds(self):

        # test
        response = S3Client().upload_file(filename="testdt01.csv", path="./data/")
        self.assertTrue(response)

    # @Test
    def test_upload_bucket_notFound_fails(self):

        # test
        try:
            S3Client().upload_file(filename="testdt01.csv", path="./data/", bucket="invalidbucket")
        except ConnectionRefusedError:
            pass
        except Exception:
            self.fail('unexpected exception raised')
        else:
            self.fail('ConnectionRefusedError not raised')

    # @Test
    def test_upload_file_notFound_fails(self):

        # test
        try:
            S3Client().upload_file(filename="testdt06.csv", path="./data/")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail('unexpected exception raised')
        else:
            self.fail('FileNotFoundError not raised')

    # @Test
    def test_download_file_succeeds(self):

        # mock
        S3Client().upload_file(filename="testdt01.csv", path="./data/", bucket="dbrepo")

        # test
        S3Client().download_file(filename="testdt01.csv", bucket="dbrepo")

    # @Test
    def test_download_file_notFound_fails(self):

        # test
        try:
            S3Client().download_file(filename="testdt01.csv", bucket="dbrepo")
        except ClientError:
            pass
        except Exception:
            self.fail('unexpected exception raised')
        else:
            self.fail('ClientError not raised')

    # @Test
    def test_download_bucket_notFound_fails(self):

        # test
        try:
            S3Client().download_file(filename="testdt01.csv", bucket="invalidbucket")
        except ClientError:
            pass
        except Exception:
            self.fail('unexpected exception raised')
        else:
            self.fail('ClientError not raised')

    # @Test
    def test_get_file_succeeds(self):

        # mock
        S3Client().upload_file(filename="testdt01.csv", path="./data/", bucket="dbrepo")

        # test
        response = S3Client().get_file(bucket="dbrepo", filename="testdt01.csv")
        self.assertIsNotNone(response)

    # @Test
    def test_get_file_notFound_fails(self):

        # test
        try:
            S3Client().get_file(bucket="dbrepo", filename="idonotexist.csv")
        except ClientError:
            pass
        except Exception:
            self.fail('unexpected exception raised')
        else:
            self.fail('ClientError not raised')

    # @Test
    def test_bucket_exists_succeeds(self):

        # test
        response = S3Client().bucket_exists_or_exit("dbrepo")
        self.assertIsNotNone(response)

    # @Test
    def test_bucket_exists_notExists_fails(self):

        # test
        try:
            S3Client().bucket_exists_or_exit("idnonotexist")
        except FileNotFoundError:
            pass
        except Exception:
            self.fail('unexpected exception raised')
        else:
            self.fail('FileNotFoundError not raised')


if __name__ == '__main__':
    unittest.main()



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import Identifier, IdentifierType, SaveIdentifierTitle, Creator, IdentifierTitle, \
    IdentifierDescription, SaveIdentifierDescription, Language, SaveIdentifierFunder, SaveRelatedIdentifier, \
    RelatedIdentifierRelation, RelatedIdentifierType, IdentifierFunder, RelatedIdentifier, UserBrief, \
    IdentifierStatusType, CreateIdentifierCreator, CreateIdentifierTitle, CreateIdentifierFunder, \
    CreateRelatedIdentifier, CreateIdentifierDescription, SaveIdentifierCreator, Links
from dbrepo.api.exceptions import MalformedError, ForbiddenError, NotExistsError, AuthenticationError, \
    ServiceConnectionError, ServiceError, ResponseCodeError, FormatNotAvailable, RequestError


class IdentifierUnitTest(unittest.TestCase):

    def test_create_identifier_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                             links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                             publication_year=2024,
                             publisher='TU Wien',
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                 description='Test Description')],
                             titles=[IdentifierTitle(id="ffaf9e0c-c37d-4655-bd68-80cd991cf24c",
                                                     title='Test Title')],
                             funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                       funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                   value='10.12345/abc',
                                                   relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[Creator(id="70539dff-c549-4c95-8257-9c750decf232",
                                               creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.post('/api/v1/identifier', json=exp.model_dump(), status_code=201)
            # test
            client = RestClient(username="a", password="b")
            response = client.create_identifier(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                titles=[CreateIdentifierTitle(title='Test Title')],
                publisher='TU Wien', publication_year=2024,
                language=Language.EN,
                funders=[CreateIdentifierFunder(funder_name='FWF')],
                related_identifiers=[CreateRelatedIdentifier(value='10.12345/abc',
                                                             relation=RelatedIdentifierRelation.CITES,
                                                             type=RelatedIdentifierType.DOI)],
                descriptions=[CreateIdentifierDescription(description='Test Description')],
                creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            self.assertEqual(exp, response)

    def test_create_identifier_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/identifier', status_code=400)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_identifier(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except MalformedError:
                pass

    def test_create_identifier_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/identifier', status_code=403)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_identifier(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except ForbiddenError:
                pass

    def test_create_identifier_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/identifier', status_code=404)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_identifier(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except NotExistsError:
                pass

    def test_create_identifier_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/identifier', status_code=502)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_identifier(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except ServiceConnectionError:
                pass

    def test_create_identifier_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/identifier', status_code=503)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_identifier(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except ServiceError:
                pass

    def test_create_identifier_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.post('/api/v1/identifier', status_code=200)
            # test
            try:
                client = RestClient(username="a", password="b")
                client.create_identifier(
                    database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                    titles=[CreateIdentifierTitle(title='Test Title')],
                    descriptions=[CreateIdentifierDescription(description='Test')],
                    publisher='TU Wien', publication_year=2024,
                    creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
            except ResponseCodeError:
                pass

    def test_create_identifier_anonymous_fails(self):
        # test
        try:
            RestClient().create_identifier(
                database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                titles=[CreateIdentifierTitle(title='Test Title')],
                descriptions=[CreateIdentifierDescription(description='Test')],
                publisher='TU Wien', publication_year=2024,
                creators=[CreateIdentifierCreator(creator_name='Carberry, Josiah')])
        except AuthenticationError:
            pass

    def test_get_identifiers_view_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                              database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                              view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                              publication_year=2024,
                              publisher='TU Wien',
                              type=IdentifierType.VIEW,
                              language=Language.EN,
                              links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                          self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                          data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                              descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                  description='Test Description')],
                              titles=[IdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                      title='Test Title')],
                              funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                        funder_name='FWF')],
                              related_identifiers=[
                                  RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815", value='10.12345/abc',
                                                    relation=RelatedIdentifierRelation.CITES,
                                                    type=RelatedIdentifierType.DOI)],
                              creators=[Creator(id="70539dff-c549-4c95-8257-9c750decf232",
                                                creator_name='Carberry, Josiah')],
                              status=IdentifierStatusType.PUBLISHED,
                              owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))]
            # mock
            mock.get('/api/v1/identifier', json=[exp[0].model_dump()], headers={"Accept": "application/json"})
            # test
            response = RestClient().get_identifiers(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                    view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                                                    type=IdentifierType.VIEW,
                                                    status=IdentifierStatusType.PUBLISHED)
            self.assertEqual(exp, response)

    def test_get_identifiers_subset_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = []
            # mock
            mock.get('/api/v1/identifier', json=[], headers={"Accept": "application/json"})
            # test
            response = RestClient().get_identifiers(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                    subset_id="0831bf54-9dd9-46fe-8c2c-c539332ea177")
            self.assertEqual(exp, response)

    def test_get_identifiers_table_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = []
            # mock
            mock.get('/api/v1/identifier', json=[], headers={"Accept": "application/json"})
            # test
            response = RestClient().get_identifiers(database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                    table_id="b3230b86-4743-498d-9015-3fad58049692")
            self.assertEqual(exp, response)

    def test_get_identifiers_view_param_database_fails(self):
        # test
        try:
            RestClient().get_identifiers(view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c")
        except RequestError:
            pass

    def test_get_identifiers_subset_param_database_fails(self):
        # test
        try:
            RestClient().get_identifiers(subset_id="0831bf54-9dd9-46fe-8c2c-c539332ea177")
        except RequestError:
            pass

    def test_get_identifiers_table_param_database_fails(self):
        # test
        try:
            RestClient().get_identifiers(table_id="b3230b86-4743-498d-9015-3fad58049692")
        except RequestError:
            pass

    def test_get_identifiers_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/identifier', status_code=404)
            # test
            try:
                RestClient().get_identifiers()
            except NotExistsError:
                pass

    def test_get_identifiers_406_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/identifier', status_code=406)
            # test
            try:
                RestClient().get_identifiers()
            except FormatNotAvailable:
                pass

    def test_get_identifiers_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/identifier', status_code=202)
            # test
            try:
                RestClient().get_identifiers()
            except ResponseCodeError:
                pass

    def test_update_identifier_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                             publication_year=2024,
                             publisher='TU Wien',
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                             descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                 description='Test Description')],
                             titles=[IdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                     title='Test Title')],
                             funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                       funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                   value='10.12345/abc',
                                                   relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[Creator(id="70539dff-c549-4c95-8257-9c750decf232",
                                               creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3', json=exp.model_dump(), status_code=202)
            # test
            client = RestClient(username="a", password="b")
            response = client.update_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                                                type=IdentifierType.VIEW,
                                                titles=[SaveIdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                                            title='Test Title')],
                                                publisher='TU Wien', publication_year=2024,
                                                language=Language.EN,
                                                funders=[SaveIdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                                              funder_name='FWF')],
                                                related_identifiers=[
                                                    SaveRelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                                          value='10.12345/abc',
                                                                          relation=RelatedIdentifierRelation.CITES,
                                                                          type=RelatedIdentifierType.DOI)],
                                                descriptions=[
                                                    SaveIdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                              description='Test Description')],
                                                creators=[
                                                    SaveIdentifierCreator(id="6bf894bc-8f55-4b5d-83cf-198b29253260",
                                                                          creator_name='Carberry, Josiah')])
            self.assertEqual(exp, response)

    def test_update_identifier_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3', status_code=400)
            # test
            client = RestClient(username="a", password="b")
            try:
                client.update_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                                         titles=[SaveIdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                                     title='Test Title')],
                                         publisher='TU Wien', publication_year=2024,
                                         language=Language.EN,
                                         funders=[SaveIdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                                       funder_name='FWF')],
                                         related_identifiers=[
                                             SaveRelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                                   value='10.12345/abc',
                                                                   relation=RelatedIdentifierRelation.CITES,
                                                                   type=RelatedIdentifierType.DOI)],
                                         descriptions=[
                                             SaveIdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                       description='Test Description')],
                                         creators=[SaveIdentifierCreator(id="6bf894bc-8f55-4b5d-83cf-198b29253260",
                                                                         creator_name='Carberry, Josiah')])
            except MalformedError:
                pass

    def test_update_identifier_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3', status_code=403)
            # test
            client = RestClient(username="a", password="b")
            try:
                client.update_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                                         titles=[SaveIdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                                     title='Test Title')],
                                         publisher='TU Wien', publication_year=2024,
                                         language=Language.EN,
                                         funders=[SaveIdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                                       funder_name='FWF')],
                                         related_identifiers=[
                                             SaveRelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                                   value='10.12345/abc',
                                                                   relation=RelatedIdentifierRelation.CITES,
                                                                   type=RelatedIdentifierType.DOI)],
                                         descriptions=[
                                             SaveIdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                       description='Test Description')],
                                         creators=[SaveIdentifierCreator(id="6bf894bc-8f55-4b5d-83cf-198b29253260",
                                                                         creator_name='Carberry, Josiah')])
            except ForbiddenError:
                pass

    def test_update_identifier_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3', status_code=404)
            # test
            client = RestClient(username="a", password="b")
            try:
                client.update_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                                         titles=[SaveIdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                                     title='Test Title')],
                                         publisher='TU Wien', publication_year=2024,
                                         language=Language.EN,
                                         funders=[SaveIdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                                       funder_name='FWF')],
                                         related_identifiers=[
                                             SaveRelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                                   value='10.12345/abc',
                                                                   relation=RelatedIdentifierRelation.CITES,
                                                                   type=RelatedIdentifierType.DOI)],
                                         descriptions=[
                                             SaveIdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                       description='Test Description')],
                                         creators=[SaveIdentifierCreator(id="6bf894bc-8f55-4b5d-83cf-198b29253260",
                                                                         creator_name='Carberry, Josiah')])
            except NotExistsError:
                pass

    def test_update_identifier_502_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3', status_code=502)
            # test
            client = RestClient(username="a", password="b")
            try:
                client.update_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                                         titles=[SaveIdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                                     title='Test Title')],
                                         publisher='TU Wien', publication_year=2024,
                                         language=Language.EN,
                                         funders=[SaveIdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                                       funder_name='FWF')],
                                         related_identifiers=[
                                             SaveRelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                                   value='10.12345/abc',
                                                                   relation=RelatedIdentifierRelation.CITES,
                                                                   type=RelatedIdentifierType.DOI)],
                                         descriptions=[
                                             SaveIdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                       description='Test Description')],
                                         creators=[SaveIdentifierCreator(id="6bf894bc-8f55-4b5d-83cf-198b29253260",
                                                                         creator_name='Carberry, Josiah')])
            except ServiceConnectionError:
                pass

    def test_update_identifier_503_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3', status_code=503)
            # test
            client = RestClient(username="a", password="b")
            try:
                client.update_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                                         titles=[SaveIdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                                     title='Test Title')],
                                         publisher='TU Wien', publication_year=2024,
                                         language=Language.EN,
                                         funders=[SaveIdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                                       funder_name='FWF')],
                                         related_identifiers=[
                                             SaveRelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                                   value='10.12345/abc',
                                                                   relation=RelatedIdentifierRelation.CITES,
                                                                   type=RelatedIdentifierType.DOI)],
                                         descriptions=[
                                             SaveIdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                       description='Test Description')],
                                         creators=[SaveIdentifierCreator(id="6bf894bc-8f55-4b5d-83cf-198b29253260",
                                                                         creator_name='Carberry, Josiah')])
            except ServiceError:
                pass

    def test_update_identifier_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3', status_code=200)
            # test
            client = RestClient(username="a", password="b")
            try:
                client.update_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                                         titles=[SaveIdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                                     title='Test Title')],
                                         publisher='TU Wien', publication_year=2024,
                                         language=Language.EN,
                                         funders=[SaveIdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                                       funder_name='FWF')],
                                         related_identifiers=[
                                             SaveRelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                                   value='10.12345/abc',
                                                                   relation=RelatedIdentifierRelation.CITES,
                                                                   type=RelatedIdentifierType.DOI)],
                                         descriptions=[
                                             SaveIdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                       description='Test Description')],
                                         creators=[SaveIdentifierCreator(id="6bf894bc-8f55-4b5d-83cf-198b29253260",
                                                                         creator_name='Carberry, Josiah')])
            except ResponseCodeError:
                pass

    def test_update_identifier_anonymous_fails(self):
        # test
        try:
            RestClient().update_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                           database_id="6bd39359-b154-456d-b9c2-caa516a45732", type=IdentifierType.VIEW,
                                           titles=[SaveIdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                                                       title='Test Title')],
                                           publisher='TU Wien', publication_year=2024,
                                           language=Language.EN,
                                           funders=[SaveIdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c",
                                                                         funder_name='FWF')],
                                           related_identifiers=[
                                               SaveRelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815",
                                                                     value='10.12345/abc',
                                                                     relation=RelatedIdentifierRelation.CITES,
                                                                     type=RelatedIdentifierType.DOI)],
                                           descriptions=[
                                               SaveIdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                         description='Test Description')],
                                           creators=[SaveIdentifierCreator(id="6bf894bc-8f55-4b5d-83cf-198b29253260",
                                                                           creator_name='Carberry, Josiah')])
        except AuthenticationError:
            pass

    def test_publish_identifier_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                             publication_year=2024,
                             publisher='TU Wien',
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                             descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                 description='Test Description')],
                             titles=[IdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3", title='Test Title')],
                             funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c", funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815", value='10.12345/abc',
                                                   relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[
                                 Creator(id="70539dff-c549-4c95-8257-9c750decf232", creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3/publish', json=exp.model_dump(),
                     status_code=202)
            # test
            client = RestClient(username="a", password="b")
            response = client.publish_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3")
            self.assertEqual(exp, response)

    def test_publish_identifier_400_fails(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                             publication_year=2024,
                             publisher='TU Wien',
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                             descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                 description='Test Description')],
                             titles=[IdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3", title='Test Title')],
                             funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c", funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815", value='10.12345/abc',
                                                   relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[
                                 Creator(id="70539dff-c549-4c95-8257-9c750decf232", creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3/publish', json=exp.model_dump(),
                     status_code=400)
            # test
            try:
                RestClient(username="a", password="b").publish_identifier(
                    identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3")
            except MalformedError:
                pass

    def test_publish_identifier_403_fails(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                             publication_year=2024,
                             publisher='TU Wien',
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                             descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                 description='Test Description')],
                             titles=[IdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3", title='Test Title')],
                             funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c", funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815", value='10.12345/abc',
                                                   relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[
                                 Creator(id="70539dff-c549-4c95-8257-9c750decf232", creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3/publish', json=exp.model_dump(),
                     status_code=403)
            # test
            try:
                RestClient(username="a", password="b").publish_identifier(
                    identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3")
            except ForbiddenError:
                pass

    def test_publish_identifier_404_fails(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                             publication_year=2024,
                             publisher='TU Wien',
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                             descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                 description='Test Description')],
                             titles=[IdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3", title='Test Title')],
                             funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c", funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815", value='10.12345/abc',
                                                   relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[
                                 Creator(id="70539dff-c549-4c95-8257-9c750decf232", creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3/publish', json=exp.model_dump(),
                     status_code=404)
            # test
            try:
                RestClient(username="a", password="b").publish_identifier(
                    identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3")
            except NotExistsError:
                pass

    def test_publish_identifier_502_fails(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                             publication_year=2024,
                             publisher='TU Wien',
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                             descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                 description='Test Description')],
                             titles=[IdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3", title='Test Title')],
                             funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c", funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815", value='10.12345/abc',
                                                   relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[
                                 Creator(id="70539dff-c549-4c95-8257-9c750decf232", creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3/publish', json=exp.model_dump(),
                     status_code=502)
            # test
            try:
                RestClient(username="a", password="b").publish_identifier(
                    identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3")
            except ServiceConnectionError:
                pass

    def test_publish_identifier_503_fails(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                             publication_year=2024,
                             publisher='TU Wien',
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                             descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                 description='Test Description')],
                             titles=[IdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3", title='Test Title')],
                             funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c", funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815", value='10.12345/abc',
                                                   relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[
                                 Creator(id="70539dff-c549-4c95-8257-9c750decf232", creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3/publish', json=exp.model_dump(),
                     status_code=503)
            # test
            try:
                RestClient(username="a", password="b").publish_identifier(
                    identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3")
            except ServiceError:
                pass

    def test_publish_identifier_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            exp = Identifier(id="f6171539-a479-4829-9b9b-a6b474e1c7d3",
                             database_id="6bd39359-b154-456d-b9c2-caa516a45732",
                             view_id="e5229d24-584a-43e8-b9f6-d349c3053f9c",
                             publication_year=2024,
                             publisher='TU Wien',
                             type=IdentifierType.VIEW,
                             language=Language.EN,
                             links=Links(self="/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         self_html="/pid/f6171539-a479-4829-9b9b-a6b474e1c7d3",
                                         data="/api/v1/database/6bd39359-b154-456d-b9c2-caa516a45732/view/e5229d24-584a-43e8-b9f6-d349c3053f9c/data"),
                             descriptions=[IdentifierDescription(id="d8bdc933-655c-46bd-9903-ede3928a304b",
                                                                 description='Test Description')],
                             titles=[IdentifierTitle(id="f6171539-a479-4829-9b9b-a6b474e1c7d3", title='Test Title')],
                             funders=[IdentifierFunder(id="d0dc801d-cfb7-4b07-9b20-2d7af39c913c", funder_name='FWF')],
                             related_identifiers=[
                                 RelatedIdentifier(id="6655eba7-b0ac-4bc4-9f09-6355fec8d815", value='10.12345/abc',
                                                   relation=RelatedIdentifierRelation.CITES,
                                                   type=RelatedIdentifierType.DOI)],
                             creators=[
                                 Creator(id="70539dff-c549-4c95-8257-9c750decf232", creator_name='Carberry, Josiah')],
                             status=IdentifierStatusType.PUBLISHED,
                             owner=UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise'))
            # mock
            mock.put('/api/v1/identifier/f6171539-a479-4829-9b9b-a6b474e1c7d3/publish', json=exp.model_dump(),
                     status_code=200)
            # test
            try:
                RestClient(username="a", password="b").publish_identifier(
                    identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3")
            except ResponseCodeError:
                pass

    def test_publish_identifier_anonymous_fails(self):
        # test
        try:
            RestClient().publish_identifier(identifier_id="f6171539-a479-4829-9b9b-a6b474e1c7d3")
        except AuthenticationError:
            pass


if __name__ == "__main__":
    unittest.main()



import logging
from typing import List

import requests
from grafana_client import GrafanaApi
from grafana_client.client import GrafanaClientError, GrafanaException
from requests import Response

from dbrepo.api.dto import Database, ColumnType, ViewColumn, View
from dbrepo.core.api.dto import Permission

statistics_row_title = 'Generated Dashboard'
auto_generated_description = 'Auto-generated'

number_types = [ColumnType.SERIAL, ColumnType.BIT, ColumnType.SMALLINT, ColumnType.MEDIUMINT, ColumnType.INT,
                ColumnType.BIGINT, ColumnType.FLOAT, ColumnType.DOUBLE, ColumnType.DECIMAL]

time_types = [ColumnType.DATE, ColumnType.TIME, ColumnType.TIMESTAMP, ColumnType.YEAR]

bool_types = [ColumnType.TINYINT, ColumnType.BOOL]

section_height = 3 * 8


def map_link(title: str, url: str, icon: str = 'info', open_new_window: bool = True) -> dict:
    return dict(targetBlank=open_new_window,
                asDropdown=False,
                includeVars=False,
                keepTime=False,
                tags=[],
                type='link',
                icon=icon,
                title=title,
                url=url)


def _get_managed_offset_y(dashboard: dict) -> int | None:
    idx = [panel['title'] for panel in dashboard['panels']].index(statistics_row_title)
    if idx == -1:
        return None
    offset_y = dashboard['panels'][idx]['gridPos']['y']
    logging.debug(f'managed panel y-offset: {offset_y}')
    return offset_y


def _get_start_index(dashboard: dict) -> int | None:
    return [panel['title'] for panel in dashboard['panels']].index(statistics_row_title)


def map_column_conversion(column: ViewColumn) -> dict:
    destinationType = 'string'
    dateFormat = None
    if column.type in number_types:
        destinationType = 'number'
    elif column.type in time_types:
        destinationType = 'time'
        if column.type == ColumnType.YEAR:
            dateFormat = 'YYYY'
        elif column.type == ColumnType.TIME:
            dateFormat = 'HH:mm:ss'
        else:
            dateFormat = 'YYYY-MM-dd'
    elif column.type in bool_types:
        destinationType = 'boolean'
    return dict(targetField=column.internal_name,
                destinationType=destinationType,
                dateFormat=dateFormat)


def map_row(title: str, x: int = 0, y: int = 0) -> dict:
    return dict(collapsed=False,
                title=title,
                type='row',
                panels=[],
                targets=[],
                parser='backend',
                gridPos=dict(h=1,
                             w=24,
                             x=x,
                             y=y))


def map_preview_image_panel(database_id: str, w: int = 4, h: int = 4, x: int = 20, y: int = 0) -> dict:
    return dict(title='Preview Image',
                type='text',
                description=auto_generated_description,
                gridPos=dict(h=h,
                             w=w,
                             x=x,
                             y=y),
                fieldConfig=dict(defaults=dict(),
                                 overrides=[]),
                options=dict(mode="markdown",
                             code=dict(language="plaintext",
                                       showLineNumbers=False,
                                       showMiniMap=False),
                             content=f'<img src="/api/v1/database/{database_id}/image" alt="" width="90" />'))


class DashboardServiceClient:

    def __init__(self, endpoint: str, username: str, password: str, base_url: str = 'http://localhost',
                 datasource_uid: str = 'dbrepojson0'):
        self.client: GrafanaApi = GrafanaApi.from_url(url=f'{endpoint}', credential=(username, password))
        self.endpoint = endpoint
        self.username = username
        self.password = password
        self.base_url = base_url
        self.datasource_uid = datasource_uid

    def get_client(self):
        return self.client

    def generic_get(self, api_url: str) -> Response:
        request_url = self.endpoint + api_url
        logging.debug(f'generic get url={request_url}, auth=({self.username}, <reacted>)')
        return requests.get(request_url, auth=(self.username, self.password))

    def generic_post(self, api_url: str, payload: dict) -> Response:
        request_url = self.endpoint + api_url
        logging.debug(f'generic post url={request_url}, payload={payload}, auth=({self.username}, <reacted>)')
        return requests.post(request_url, json=payload, auth=(self.username, self.password))

    def find(self, uid: str):
        """
        Finds a dashboard with the given uid.

        @return The dashboard, if successful. Otherwise, `None`.
        """
        if uid is None:
            return None
        try:
            return self.client.dashboard.get_dashboard(uid)
        except GrafanaClientError:
            logging.warning(f"Failed to find dashboard with uid: {uid}")
            return None

    def create(self, database_name: str, uid: str = '') -> dict:
        dashboard = dict(uid=uid,
                         title=f'{database_name} Overview',
                         tags=['managed'],
                         timezone='browser',
                         refresh='30m',
                         preload=False,
                         panels=[])
        dashboard['panels'] = []
        payload = dict(folderUid='',
                       overwrite=False,
                       dashboard=dashboard)
        dashboard = self.client.dashboard.update_dashboard(payload)
        logging.info(f"Created dashboard with uid: {dashboard['uid']}")
        return dashboard

    def delete(self, uid: str) -> None:
        self.client.dashboard.delete_dashboard(uid)

    def update(self, database: Database) -> None:
        dashboard = self.find(database.dashboard_uid)
        if dashboard is None:
            self.create(database.internal_name, database.dashboard_uid)
            dashboard = self.find(database.dashboard_uid)
        dashboard = dashboard['dashboard']
        # update metadata
        if not database.is_dashboard_enabled and 'managed' in dashboard['tags']:
            dashboard['tags'].remove('managed')
        if len(database.identifiers) > 0 and len(database.identifiers[0].titles) > 0:
            dashboard['title'] = database.identifiers[0].titles[0].title
        if len(database.identifiers) > 0 and len(database.identifiers[0].descriptions) > 0:
            dashboard['description'] = database.identifiers[0].descriptions[0].description
        dashboard['links'] = self.map_links(database)
        # update panels
        dashboard['panels'] = self.get_panels(dashboard, database)
        payload = dict(folderUid='',
                       overwrite=True,
                       dashboard=dashboard)
        response = self.client.dashboard.update_dashboard(payload)
        logging.info(f"Updated dashboard with uid: {response['uid']}")

    def map_links(self, database: Database) -> List[dict]:
        links = []
        if len(database.identifiers) > 0:
            links.append(map_link('Database', f"{self.base_url}/pid/{database.identifiers[0].id}"))
        else:
            links.append(map_link('Database', f"{self.base_url}/database/{database.id}"))
        return links

    def update_anonymous_read_access(self, uid: str, is_public: bool, is_schema_public: bool) -> None:
        permissions = self.client.dashboard.get_permissions_by_uid(uid)
        viewer_role = [permission for permission in permissions if
                       'permissionName' in permission and permission['permissionName'] != 'View']
        permission = ''
        if is_public or is_schema_public:
            permission = 'View'
        if len(viewer_role) == 0:
            logging.warning(f'Failed to find permissionName=View')
            return None
        try:
            response = self.generic_post(f'/api/access-control/dashboards/{uid}/builtInRoles/Viewer',
                                         Permission(permission=permission).model_dump())
            if response.status_code != 200:
                raise OSError(f'Failed to update anonymous read access: {response.content}')
        except GrafanaException as e:
            raise OSError(f'Failed to update anonymous read access: {e.message}')
        logging.info(f"Updated anonymous read access for dashboard with uid: {uid}")

    def _map_timeseries_panel(self, database_id: str, view: View, panel_type: str, h: int = 8, w: int = 12, x: int = 12,
                              y: int = 8) -> dict:
        datasource = dict(uid=self.datasource_uid,
                          type='yesoreyeram-infinity-datasource')
        fillOpacity = 0
        if panel_type == 'histogram':
            fillOpacity = 60
        return dict(title=panel_type.capitalize(),
                    description=auto_generated_description,
                    type=panel_type,
                    datasource=datasource,
                    targets=[dict(datasource=datasource,
                                  format='table',
                                  global_query_id='',
                                  hide=False,
                                  refId='A',
                                  root_selector='',
                                  source='url',
                                  type='json',
                                  url=f'/api/v1/database/{database_id}/view/{view.id}/data',
                                  parser='backend',
                                  url_options=dict(data='',
                                                   method='GET'))],
                    gridPos=dict(h=h,
                                 w=w,
                                 x=x,
                                 y=y),
                    options=dict(legend=dict(displayMode='list',
                                             placement='bottom',
                                             showLegend=True),
                                 tooltip=dict(mode='single',
                                              sort='none')),
                    fieldConfig=dict(
                        defaults=dict(color=dict(mode='palette-classic'),
                                      custom=dict(
                                          axisBorderShow=False,
                                          axisCenteredZero=False,
                                          axisColorMode='text',
                                          axisLabel='',
                                          axisPlacement='auto',
                                          barAlignment=0,
                                          drawStyle='line',
                                          fillOpacity=fillOpacity,
                                          gradientMode='none',
                                          hideFrom=dict(legend=False,
                                                        tooltip=False,
                                                        viz=False),
                                          insertNulls=False,
                                          lineInterpolation='linear',
                                          lineWidth=1,
                                          pointSize=5,
                                          scaleDistribution=dict(type='linear'),
                                          showPoints='auto',
                                          spanNulls=False,
                                          stacking=dict(group='A',
                                                        mode='none'),
                                          thresholdsStyle=dict(mode='absolute')))),
                    transformations=[dict(id='convertFieldType',
                                          options=dict(fields=dict(),
                                                       conversions=[map_column_conversion(column) for column in
                                                                    view.columns]))])

    def _map_number_panel(self, database_id: str, view_id: str, title: str, field: str, x: int = 18,
                          y: int = 0) -> dict:
        datasource = dict(uid=self.datasource_uid,
                          type='yesoreyeram-infinity-datasource')
        return dict(title=title,
                    type='stat',
                    datasource=datasource,
                    targets=[dict(datasource=datasource,
                                  columns=[],
                                  filters=[],
                                  format='table',
                                  global_query_id='',
                                  hide=False,
                                  refId='A',
                                  root_selector='',
                                  source='url',
                                  type='json',
                                  url=f'/api/v1/database/{database_id}/view/{view_id}/statistic',
                                  parser='backend',
                                  url_options=dict(data='',
                                                   method='GET'))],
                    fieldConfig=dict(defaults=dict(mappings=[],
                                                   thresholds=dict(mode='absolute',
                                                                   steps=[dict(color='blue',
                                                                               value=None)]),
                                                   unit=''),
                                     overrides=[]),
                    transformations=[dict(id='extractFields',
                                          options=dict(delimiter=',',
                                                       source=field,
                                                       format='auto',
                                                       replace=False,
                                                       keepTime=False)),
                                     dict(id='filterFieldsByName',
                                          options=dict(include=dict(names=[field])))],
                    gridPos=dict(h=4,
                                 w=6,
                                 x=x,
                                 y=y),
                    options=dict(colorMode='background',
                                 graphMode='area',
                                 justifyMode='auto',
                                 orientation='auto',
                                 reduceOptions=dict(calcs=[],
                                                    fields='/.*/',
                                                    values=True),
                                 showPercentChange=False,
                                 textMode='auto',
                                 wideLayout=True))

    def map_overview_panel(self, database_id: str, view_id: str, x: int = 0, y: int = 4) -> dict:
        datasource = dict(uid=self.datasource_uid,
                          type='yesoreyeram-infinity-datasource')
        return dict(title='Datasource Preview',
                    type='table',
                    gridPos=dict(h=8,
                                 w=18,
                                 x=x,
                                 y=y),
                    fieldConfig=dict(
                        defaults=dict(
                            color=dict(mode='palette-classic'),
                            custom=dict(axisBorderShow=False,
                                        axisCenteredZero=False,
                                        axisColorMode='text',
                                        axisLabel='',
                                        axisPlacement='auto',
                                        barAlignment=0,
                                        drawStyle='line',
                                        fillOpacity=0,
                                        gradientMode='none',
                                        hideFrom=dict(
                                            legend=False,
                                            tooltip=False,
                                            viz=False),
                                        insertNulls=False,
                                        lineInterpolation='linear',
                                        lineWidth=1,
                                        pointSize=5,
                                        scaleDistribution=dict(
                                            type='linear'),
                                        showPoints='auto',
                                        spanNulls=False,
                                        stacking=dict(group='A',
                                                      mode='none'),
                                        thresholdsStyle=dict(
                                            mode='off'))),
                        overrides=[]),
                    options=dict(legend=dict(displayMode='list',
                                             placement='bottom',
                                             showLegend=True,
                                             calcs=[]),
                                 tooltip=dict(mode='single',
                                              sort='none')),
                    targets=[dict(format='json',
                                  columns=[],
                                  datasource=datasource,
                                  filters=[],
                                  global_query_id='',
                                  refId='A',
                                  root_selector='',
                                  source='url',
                                  type='json',
                                  url=f'/api/v1/database/{database_id}/view/{view_id}/data',
                                  parser='backend',
                                  url_options=dict(data='',
                                                   method='GET'))],
                    links=[dict(title='Cite',
                                url=f'{self.base_url}/database/{database_id}/view/{view_id}/data',
                                targetBlank=True)],
                    datasource=datasource)

    def map_statistics_panel(self, database_id: str, view_id: str, w: int = 12, h: int = 8, x: int = 0,
                             y: int = 8) -> dict:
        datasource = dict(uid=self.datasource_uid,
                          type='yesoreyeram-infinity-datasource')
        return dict(title='Statistics',
                    type='table',
                    gridPos=dict(h=h,
                                 w=w,
                                 x=x,
                                 y=y),
                    datasource=datasource,
                    targets=[dict(datasource=datasource,
                                  columns=[],
                                  filters=[],
                                  format='table',
                                  global_query_id='',
                                  hide=False,
                                  refId='A',
                                  root_selector='columns',
                                  source='url',
                                  type='json',
                                  url=f'/api/v1/database/{database_id}/view/{view_id}/statistic',
                                  parser='backend',
                                  url_options=dict(data='',
                                                   method='GET'))],
                    options=dict(cellHeight="sm",
                                 showHeader=True,
                                 footer=dict(countRows=False,
                                             fields="",
                                             reducer=["sum"],
                                             show=False)),
                    transformations=[dict(id="organize",
                                          options=dict(excludeByName=dict(),
                                                       includeByName=dict(),
                                                       indexByName=dict(name=0,
                                                                        val_min=1,
                                                                        val_max=2,
                                                                        mean=3,
                                                                        median=4,
                                                                        std_dev=5),
                                                       renameByName=dict(name="Name",
                                                                         mean="Mean",
                                                                         median="Median",
                                                                         std_dev="std.dev",
                                                                         val_min="Minimum",
                                                                         val_max="Maximum")))],
                    fieldConfig=dict(defaults=dict(custom=dict(align="auto",
                                                               filterable="true",
                                                               cellOptions=dict(type="auto"),
                                                               inspect=False),
                                                   mappings=[],
                                                   thresholds=dict(mode="absolute",
                                                                   steps=[dict(color="green",
                                                                               value=None),
                                                                          dict(color="red",
                                                                               value=80)
                                                                          ])),
                                     overrides=[]))

    def map_timeseries_panel(self, database_id: str, view: View, h: int = 8, w: int = 12, x: int = 12,
                             y: int = 8) -> dict:
        return self._map_timeseries_panel(database_id, view, 'timeseries', h, w, x, y)

    def map_pie_panel(self, database_id: str, view: View, h: int = 8, w: int = 12, x: int = 12, y: int = 8) -> dict:
        return self._map_timeseries_panel(database_id, view, 'piechart', h, w, x, y)

    def map_histogram_panel(self, database_id: str, view: View, h: int = 8, w: int = 12, x: int = 12,
                            y: int = 8) -> dict:
        return self._map_timeseries_panel(database_id, view, 'histogram', h, w, x, y)

    def map_data_sources_panel(self, database_id: str, x: int = 0, y: int = 0) -> dict:
        datasource = dict(uid=self.datasource_uid,
                          type='yesoreyeram-infinity-datasource')
        return dict(title='Datasources',
                    description=auto_generated_description,
                    type='stat',
                    datasource=datasource,
                    targets=[dict(datasource=datasource,
                                  columns=[],
                                  filters=[],
                                  format='table',
                                  global_query_id='',
                                  hide=False,
                                  refId='A',
                                  root_selector='$count(id)',
                                  source='url',
                                  type='json',
                                  url=f'/api/v1/database/{database_id}/view',
                                  parser='backend',
                                  url_options=dict(data='',
                                                   method='GET'))],
                    fieldConfig=dict(defaults=dict(mappings=[],
                                                   thresholds=dict(mode='absolute',
                                                                   steps=[dict(color='blue',
                                                                               value=None)])),
                                     overrides=[]),
                    transformations=[],
                    gridPos=dict(h=4,
                                 w=6,
                                 x=x,
                                 y=y),
                    options=dict(colorMode='background',
                                 graphMode='area',
                                 justifyMode='auto',
                                 orientation='auto',
                                 reduceOptions=dict(calcs=[],
                                                    fields='/.*/',
                                                    values=True),
                                 showPercentChange=False,
                                 textMode='auto',
                                 wideLayout=True))

    def map_rows_panel(self, database_id: str, view_id: str, x: int = 18, y: int = 0) -> dict:
        return self._map_number_panel(database_id, view_id, 'Rows', 'total_rows', x, y)

    def map_columns_panel(self, database_id: str, view_id: str, x: int = 18, y: int = 0) -> dict:
        return self._map_number_panel(database_id, view_id, 'Variables', 'total_columns', x, y)

    def get_panels(self, dashboard: dict, database: Database) -> [dict]:
        panels = dashboard['panels']
        managed_offset = 1
        try:
            managed_offset = _get_managed_offset_y(dashboard)
            end_index = _get_start_index(dashboard)
            logging.debug(f'splicing managed panels after index: {end_index}')
            panels = panels[:end_index]
        except ValueError:
            logging.warning(f'No managed panels found')
        original_panels_size = len(panels)
        panels.append(map_row(statistics_row_title, 0, managed_offset + 0))
        panels.append(self.map_data_sources_panel(database.id, y=managed_offset))
        if database.preview_image is not None:
            panels.append(map_preview_image_panel(database.id, y=managed_offset))
        for i, view in enumerate(database.views):
            # section
            panels.append(map_row(view.name, 0, y=i * section_height + managed_offset + 4))
            panels.append(self.map_overview_panel(database.id, view.id, 0, y=i * section_height + managed_offset + 8))
            panels.append(self.map_rows_panel(database.id, view.id, 18, y=i * section_height + managed_offset + 4))
            panels.append(self.map_columns_panel(database.id, view.id, 18, y=i * section_height + managed_offset + 8))
            panels.append(self.map_statistics_panel(database.id, view.id, h=8, w=12, x=0,
                                                    y=i * section_height + managed_offset + 12))
            panels.append(self.map_histogram_panel(database.id, view, h=8, w=12, x=12,
                                                   y=i * section_height + managed_offset + 12))
            panels.append(self.map_timeseries_panel(database.id, view, h=8, w=8, x=0,
                                                    y=i * section_height + managed_offset + 20))
            panels.append(self.map_pie_panel(database.id, view, h=8, w=8, x=8,
                                             y=i * section_height + managed_offset + 20))
        logging.info(f'Added {len(panels) - original_panels_size} managed panel(s)')
        return panels



import unittest

import requests_mock

from dbrepo.RestClient import RestClient
from dbrepo.api.dto import UserAttributes, UserBrief, User, UnitBrief
from dbrepo.api.exceptions import ResponseCodeError, NotExistsError, \
    ForbiddenError, AuthenticationError, ServiceError, MalformedError


class UserUnitTest(unittest.TestCase):

    def test_whoami_fails(self):
        username = RestClient().whoami()
        self.assertIsNone(username)

    def test_whoami_succeeds(self):
        client = RestClient(username="a", password="b")
        username = client.whoami()
        self.assertEqual("a", username)

    def test_get_users_empty_succeeds(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/user', json=[])
            # test
            response = RestClient().get_users()
            self.assertEqual([], response)

    def test_get_users_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = [UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise')]
            # mock
            mock.get('/api/v1/user', json=[exp[0].model_dump()])
            # test
            response = RestClient().get_users()
            self.assertEqual(exp, response)

    def test_get_users_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/user', status_code=404)
            # test
            try:
                response = RestClient().get_users()
            except ResponseCodeError as e:
                pass

    def test_get_user_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = User(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise',
                       attributes=UserAttributes(theme='dark', language='en'))
            # mock
            mock.get('/api/v1/user/user1',
                     json=exp.model_dump())
            # test
            response = RestClient().get_user(username='user1')
            self.assertEqual(exp, response)

    def test_get_user_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get('/api/v1/user/user1', status_code=404)
            # test
            try:
                response = RestClient().get_user(username='user1')
            except NotExistsError as e:
                pass

    def test_update_user_succeeds(self):
        with requests_mock.Mocker() as mock:
            exp = UserBrief(id='8638c043-5145-4be8-a3e4-4b79991b0a16', username='mweise', given_name='Martin')
            # mock
            mock.put('/api/v1/user/user1', status_code=202,
                     json=exp.model_dump())
            # test
            client = RestClient(username="a", password="b")
            response = client.update_user(username='user1', firstname='Martin',
                                          language='en', theme='light')
            self.assertEqual(exp, response)

    def test_get_user_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/user/f27921d4-b05f-4e21-a122-4953a6a779a2', status_code=403)
            # test
            try:
                RestClient().get_user('f27921d4-b05f-4e21-a122-4953a6a779a2')
            except ForbiddenError:
                pass

    def test_get_user_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/user/f27921d4-b05f-4e21-a122-4953a6a779a2', status_code=404)
            # test
            try:
                RestClient().get_user('f27921d4-b05f-4e21-a122-4953a6a779a2')
            except NotExistsError:
                pass

    def test_get_user_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.get(f'/api/v1/user/f27921d4-b05f-4e21-a122-4953a6a779a2', status_code=409)
            # test
            try:
                RestClient().get_user('f27921d4-b05f-4e21-a122-4953a6a779a2')
            except ResponseCodeError:
                pass

    def test_update_user_anonymous_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/user/f27921d4-b05f-4e21-a122-4953a6a779a2', status_code=202)
            # test
            try:
                RestClient().update_user(username='foo', theme='dark', language='en')
            except AuthenticationError:
                pass

    def test_update_user_400_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/user/foo', status_code=400)
            # test
            try:
                RestClient(username='foo', password='bar').update_user(username='foo',
                                                                       theme='dark', language='en')
            except MalformedError:
                pass

    def test_update_user_403_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/user/foo', status_code=403)
            # test
            try:
                RestClient(username='foo', password='bar').update_user(username='foo',
                                                                       theme='dark', language='en')
            except ForbiddenError:
                pass

    def test_update_user_404_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/user/foo', status_code=404)
            # test
            try:
                RestClient(username='foo', password='bar').update_user(username='foo',
                                                                       theme='dark', language='en')
            except NotExistsError:
                pass

    def test_update_user_unknown_fails(self):
        with requests_mock.Mocker() as mock:
            # mock
            mock.put('/api/v1/user/foo', status_code=200)
            # test
            try:
                RestClient(username='foo', password='bar').update_user(username='foo',
                                                                       theme='dark', language='en')
            except ResponseCodeError:
                pass


if __name__ == "__main__":
    unittest.main()
